@@ -1,11 +1,6 @@
-mod context;
-
-pub use context::{Context, Range};
-
 use std::fmt;
 
-use super::error::PolarError;
-use super::warning::PolarWarning;
+use super::{error::PolarError, sources::Context, warning::PolarWarning};
 
 #[derive(Debug)]
 pub enum Diagnostic {
@@ -13,6 +8,12 @@ pub enum Diagnostic {
     Warning(PolarWarning),
 }
 
+impl From<PolarError> for Diagnostic {
+    fn from(err: PolarError) -> Self {
+        Self::Error(err)
+    }
+}
+
 impl Diagnostic {
     pub fn is_error(&self) -> bool {
         matches!(self, Diagnostic::Error(_))
@@ -29,10 +30,9 @@ impl Diagnostic {
         };
         matches!(
             self,
-            Diagnostic::Error(PolarError {
-                kind: Parse(_) | Validation(FileLoading { .. }) | Validation(ResourceBlock { .. }),
-                ..
-            })
+            Diagnostic::Error(PolarError(
+                Parse(_) | Validation(FileLoading { .. }) | Validation(ResourceBlock { .. }),
+            ))
         )
     }
 
@@ -42,6 +42,23 @@ impl Diagnostic {
             Diagnostic::Warning(w) => w.kind(),
         }
     }
+
+    pub fn get_context(&self) -> Option<Context> {
+        match self {
+            Diagnostic::Error(e) => e.get_context(),
+            Diagnostic::Warning(w) => w.get_context(),
+        }
+    }
+}
+
+#[cfg(test)]
+impl Diagnostic {
+    pub fn unwrap_error(self) -> PolarError {
+        match self {
+            Diagnostic::Error(e) => e,
+            _ => panic!(),
+        }
+    }
 }
 
 impl fmt::Display for Diagnostic {
