@@ -5,16 +5,14 @@ use std::{
 };
 
 use crate::{
-    data_filtering::{unregistered_field_error, unsupported_op_error, PartialResults, Type},
-    error::{invalid_state_error, RuntimeError},
+    data_filtering::{PartialResults, Type},
+    error::{df_field_missing, df_unsupported_op, invalid_state, PolarResult},
     normalize::*,
     terms::*,
 };
 
 use serde::Serialize;
 
-type FilterResult<A> = core::result::Result<A, RuntimeError>;
-
 type TypeName = String;
 type FieldName = String;
 type VarName = String;
@@ -36,7 +34,7 @@ type Set<A> = HashSet<A>;
 #[derive(Clone, Eq, Debug, Serialize, PartialEq)]
 pub struct Filter {
     root: TypeName,                  // the host already has this, so we could leave it off
-    relations: Set<Relation>,        // this & root determine the "joins" (or whatever)
+    relations: Vec<Relation>,        // this & root determine the "joins" (or whatever)
     conditions: Vec<Set<Condition>>, // disjunctive normal form
 }
 
@@ -113,20 +111,20 @@ impl From<Projection> for PathVar {
 }
 
 impl PathVar {
-    fn from_term(t: &Term) -> FilterResult<Self> {
+    fn from_term(t: &Term) -> PolarResult<Self> {
         use Value::*;
         match t.value() {
             Expression(Operation {
                 operator: Operator::Dot,
                 args,
             }) => {
-                let dot = args[1].value().as_string()?.to_string();
+                let dot = args[1].as_string()?.to_string();
                 let mut pv = Self::from_term(&args[0])?;
                 pv.path.push(dot);
                 Ok(pv)
             }
             Variable(Symbol(var)) => Ok(var.clone().into()),
-            _ => invalid_state_error(format!("PathVar::from_term({})", t.to_polar())),
+            _ => invalid_state(format!("PathVar::from_term({})", t)),
         }
     }
 }
@@ -137,7 +135,7 @@ impl Filter {
         partials: PartialResults,
         var: &str,
         class: &str,
-    ) -> FilterResult<Self> {
+    ) -> PolarResult<Self> {
         let explain = std::env::var("POLAR_EXPLAIN").is_ok();
 
         if explain {
@@ -163,14 +161,14 @@ impl Filter {
         Ok(filter)
     }
 
-    fn from_partial(types: &TypeInfo, ands: Term, var: &str, class: &str) -> FilterResult<Self> {
+    fn from_partial(types: &TypeInfo, ands: Term, var: &str, class: &str) -> PolarResult<Self> {
         use {Operator::*, Value::*};
 
         if std::env::var("POLAR_EXPLAIN").is_ok() {
-            eprintln!("{}", ands.to_polar());
+            eprintln!("{}", ands);
         }
 
-        let term2expr = |i: Term| match i.value().as_expression() {
+        let term2expr = |i: Term| match i.as_expression() {
             Ok(x) => x.clone(),
             _ => op!(Unify, var!(var), i),
         };
@@ -184,7 +182,7 @@ impl Filter {
             }) => args
                 .iter()
                 .map(|and| Ok(term2expr(and.clone())))
-                .collect::<FilterResult<Vec<_>>>()
+                .collect::<PolarResult<Vec<_>>>()
                 .and_then(|ands| FilterInfo::build_filter(types.clone(), ands, var, class)),
 
             // sometimes we get an instance back. that means the variable
@@ -194,7 +192,7 @@ impl Filter {
             }
 
             // oops, we don't know how to handle this!
-            _ => invalid_state_error(ands.to_polar()),
+            _ => invalid_state(ands.to_string()),
         }
     }
 
@@ -211,10 +209,14 @@ impl Filter {
         }
     }
 
-    fn union(self, mut other: Self) -> Self {
-        other.conditions.extend(self.conditions);
-        other.relations.extend(self.relations);
-        other
+    fn union(mut self, other: Self) -> Self {
+        self.conditions.extend(other.conditions);
+        for rel in other.relations {
+            if !self.relations.iter().any(|r| r == &rel) {
+                self.relations.push(rel);
+            }
+        }
+        self
     }
 }
 
@@ -237,14 +239,14 @@ impl FilterInfo {
 
     /// turn a pathvar into a projection.
     /// populates relations as a side effect
-    fn pathvar2proj(&mut self, pv: PathVar) -> FilterResult<Projection> {
+    fn pathvar2proj(&mut self, pv: PathVar) -> PolarResult<Projection> {
         let PathVar { mut path, var } = pv;
         // new var with empty path
         let mut pv = PathVar::from(var);
         // what type is the base variable?
         let mut typ = match self.get_type(pv.clone()) {
             Some(typ) => typ,
-            _ => return invalid_state_error(format!("unknown type for `{}`", pv.var)),
+            _ => return invalid_state(format!("unknown type for `{}`", pv.var)),
         };
 
         // the last part of the path is always allowed not to be a relation.
@@ -255,7 +257,7 @@ impl FilterInfo {
         // then we fail here.
         for dot in path {
             match self.get_relation_def(&typ, &dot) {
-                None => return unregistered_field_error(&typ, &dot),
+                None => return df_field_missing(&typ, &dot),
                 Some(rel) => {
                     let Relation(_, name, right) = &rel;
                     typ = right.clone();
@@ -284,7 +286,7 @@ impl FilterInfo {
         }
     }
 
-    fn term2datum(&mut self, x: &Term) -> FilterResult<Datum> {
+    fn term2datum(&mut self, x: &Term) -> PolarResult<Datum> {
         use Datum::*;
         match PathVar::from_term(x) {
             Ok(pv) => Ok(Field(self.pathvar2proj(pv)?)),
@@ -312,32 +314,33 @@ impl FilterInfo {
     }
 
     /// Digest a conjunct from the partial results & add a new constraint.
-    fn add_constraint(&mut self, op: Operation) -> FilterResult<()> {
+    fn add_constraint(&mut self, op: Operation) -> PolarResult<()> {
         match op.args.len() {
             1 => self.add_constraint_1(op),
             2 => self.add_constraint_2(op),
-            _ => unsupported_op_error(op),
+            _ => df_unsupported_op(op),
         }
     }
 
     /// Handle a unary operation from the simplifier
-    fn add_constraint_1(&mut self, op: Operation) -> FilterResult<()> {
+    fn add_constraint_1(&mut self, op: Operation) -> PolarResult<()> {
         use Operator::*;
         // The only case this currently handles is `not in`.
         match op.operator {
-            Not => match op.args[0].value().as_expression() {
+            Not => match op.args[0].as_expression() {
                 Ok(Operation { operator: In, args }) if args.len() == 2 => {
                     let (left, right) = (self.term2datum(&args[0])?, self.term2datum(&args[1])?);
-                    self.add_condition(left, Comparison::Nin, right)
+                    self.add_condition(left, Comparison::Nin, right);
+                    Ok(())
                 }
-                _ => unsupported_op_error(op),
+                _ => df_unsupported_op(op),
             },
-            _ => unsupported_op_error(op),
+            _ => df_unsupported_op(op),
         }
     }
 
     /// Handle a binary expression from the simplifier
-    fn add_constraint_2(&mut self, op: Operation) -> FilterResult<()> {
+    fn add_constraint_2(&mut self, op: Operation) -> PolarResult<()> {
         use {Datum::*, Operator::*};
         let (left, right) = (self.term2datum(&op.args[0])?, self.term2datum(&op.args[1])?);
         let op = match op.operator {
@@ -352,28 +355,28 @@ impl FilterInfo {
             Leq => Comparison::Leq,
             Gt => Comparison::Gt,
             Geq => Comparison::Geq,
-            _ => return unsupported_op_error(op),
+            _ => return df_unsupported_op(op),
         };
-        self.add_condition(left, op, right)
+        self.add_condition(left, op, right);
+        Ok(())
     }
 
-    fn add_condition(&mut self, left: Datum, op: Comparison, right: Datum) -> FilterResult<()> {
+    fn add_condition(&mut self, left: Datum, op: Comparison, right: Datum) {
         use Comparison::*;
         match op {
-            Eq | Leq | Geq if left == right => Ok(()),
+            Eq | Leq | Geq if left == right => (),
             _ => {
                 self.conditions.insert(Condition(left, op, right));
-                Ok(())
             }
         }
     }
 
     /// Validate FilterInfo before constructing a Filter
-    fn validate(self, root: &str) -> FilterResult<Self> {
+    fn validate(self, root: &str) -> PolarResult<Self> {
         let mut set = singleton(root);
         for Relation(_, _, dst) in self.relations.iter() {
             if set.contains(dst as &str) {
-                return invalid_state_error(format!(
+                return invalid_state(format!(
                     "Type `{}` occurs more than once as the target of a relation",
                     dst
                 ));
@@ -385,7 +388,7 @@ impl FilterInfo {
     }
 
     /// populate conditions and relations on an initialized FilterInfo
-    fn with_constraints(mut self, ops: Set<Operation>, class: &str) -> FilterResult<Self> {
+    fn with_constraints(mut self, ops: Set<Operation>, class: &str) -> PolarResult<Self> {
         // find pairs of implicitly equal variables
         let equivs = ops.iter().filter_map(|Operation { operator, args }| {
             use Operator::*;
@@ -443,7 +446,27 @@ impl FilterInfo {
         parts: Vec<Operation>,
         var: &str,
         class: &str,
-    ) -> FilterResult<Filter> {
+    ) -> PolarResult<Filter> {
+        fn sort_relations(
+            relations: HashSet<Relation>,
+            mut types: HashSet<TypeName>,
+            mut out: Vec<Relation>,
+        ) -> Vec<Relation> {
+            if relations.is_empty() {
+                return out;
+            }
+            let mut rest = HashSet::new();
+            for rel in relations {
+                if types.contains(&rel.0) {
+                    types.insert(rel.2.clone());
+                    out.push(rel);
+                } else {
+                    rest.insert(rel);
+                }
+            }
+            sort_relations(rest, types, out)
+        }
+
         // TODO(gw) check more isas in host -- rn we only check external instances
         let (_isas, othas): (Set<_>, Set<_>) = parts
             .into_iter()
@@ -464,6 +487,8 @@ impl FilterInfo {
         }
         .with_constraints(othas, class)?;
 
+        let relations = sort_relations(relations, singleton(class.to_string()), vec![]);
+
         Ok(Filter {
             relations,
             conditions: vec![conditions],
@@ -534,7 +559,7 @@ impl Display for Datum {
     fn fmt(&self, f: &mut Formatter) -> Result<(), fmt::Error> {
         use Datum::*;
         match self {
-            Immediate(val) => write!(f, "{}", val.to_polar()),
+            Immediate(val) => write!(f, "{}", val),
             Field(Projection(typ, None)) => write!(f, "{}", typ),
             Field(Projection(typ, Some(field))) => write!(f, "{}.{}", typ, field),
         }
@@ -565,7 +590,7 @@ where
 fn vec_of_ands(t: Term) -> Vec<Term> {
     fn or_of_ands(t: Term) -> Vec<Term> {
         use Operator::*;
-        match t.value().as_expression() {
+        match t.as_expression() {
             Ok(Operation { operator, args }) if *operator == Or => {
                 args.iter().cloned().flat_map(or_of_ands).collect()
             }
@@ -580,7 +605,7 @@ fn vec_of_ands(t: Term) -> Vec<Term> {
 
     fn ands(t: Term) -> Vec<Term> {
         use Operator::*;
-        match t.value().as_expression() {
+        match t.as_expression() {
             Ok(Operation { operator, args }) if *operator == And => {
                 args.iter().cloned().flat_map(ands).collect()
             }
@@ -594,22 +619,25 @@ fn vec_of_ands(t: Term) -> Vec<Term> {
 #[cfg(test)]
 mod test {
     use super::*;
+    use crate::error::{ErrorKind, OperationalError, PolarError};
     use crate::events::ResultEvent;
 
-    type TestResult = Result<(), RuntimeError>;
+    type TypeMap = Map<String, Map<String, Type>>;
+
+    fn types_0() -> TypeMap {
+        let s = String::from;
+        hashmap! {
+            s("Foo") => hashmap!{
+                s("id") => Type::Base {
+                    class_tag: s("Integer")
+                },
+            }
+        }
+    }
 
     #[test]
-    fn test_or_normalization() -> TestResult {
-        let s = String::from;
-        let types = || {
-            hashmap! {
-                s("Foo") => hashmap!{
-                    s("id") => Type::Base {
-                        class_tag: s("Integer")
-                    },
-                }
-            }
-        };
+    fn test_or_normalization() -> PolarResult<()> {
+        let types = types_0;
 
         // two conditions behind an `or` in one result
         let ex1 = vec![ResultEvent::new(hashmap! {
@@ -647,10 +675,9 @@ mod test {
         Ok(())
     }
 
-    #[test]
-    fn test_dup_reln() {
+    fn types_1() -> TypeMap {
         let s = String::from;
-        let types = hashmap! {
+        hashmap! {
             s("Resource") => hashmap!{
                 s("foo") => Type::Relation {
                    kind: s("one"),
@@ -670,7 +697,12 @@ mod test {
                     other_class_tag: s("Resource"),
                 }
             }
-        };
+        }
+    }
+
+    #[test]
+    fn test_dup_reln() {
+        let types = types_1();
 
         let ors = vec![ResultEvent::new(hashmap! {
             sym!("resource") => term!(op!(And,
@@ -681,14 +713,14 @@ mod test {
         })];
 
         match Filter::build(types, ors, "resource", "Resource") {
-            Err(RuntimeError::InvalidState { msg })
+            Err(PolarError(ErrorKind::Operational(OperationalError::InvalidState { msg })))
                 if &msg == "Type `Resource` occurs more than once as the target of a relation" => {}
             x => panic!("unexpected: {:?}", x),
         }
     }
 
     #[test]
-    fn test_in() -> TestResult {
+    fn test_in() -> PolarResult<()> {
         let s = String::from;
         let types = hashmap! {
             s("Resource") => hashmap!{
@@ -724,9 +756,7 @@ mod test {
 
         assert_eq!(
             relations,
-            hashset! {
-                Relation(s("Resource"), s("foos"), s("Foo"))
-            }
+            vec![Relation(s("Resource"), s("foos"), s("Foo"))]
         );
 
         assert_eq!(
@@ -752,8 +782,67 @@ mod test {
         ];
 
         let to_s =
-            |ooa: Vec<Term>| format!("{:?}", ooa.iter().map(|a| a.to_polar()).collect::<Vec<_>>());
+            |ooa: Vec<Term>| format!("{:?}", ooa.iter().map(Term::to_string).collect::<Vec<_>>());
 
         assert_eq!(to_s(oa), to_s(vec_of_ands(ex)));
     }
+
+    fn types_2() -> TypeMap {
+        let s = String::from;
+        hashmap! {
+            s("Resource") => hashmap!{
+                s("foo") => Type::Relation {
+                   kind: s("one"),
+                   my_field: s("_"),
+                   other_field: s("_"),
+                   other_class_tag: s("Foo")
+                }
+            },
+            s("Foo") => hashmap!{
+                s("boo") => Type::Relation {
+                    kind: s("one"),
+                    my_field: s("_"),
+                    other_field: s("_"),
+                    other_class_tag: s("Boo"),
+                }
+            },
+            s("Boo") => hashmap!{
+                s("goo") => Type::Relation {
+                    kind: s("one"),
+                    my_field: s("_"),
+                    other_field: s("_"),
+                    other_class_tag: s("Goo"),
+                }
+            },
+            s("Goo") => hashmap!{
+                s("id") => Type::Base {
+                    class_tag: s("Integer")
+                }
+            }
+        }
+    }
+
+    #[test]
+    fn test_relation_depsort() -> PolarResult<()> {
+        let s = String::from;
+        let types = types_2();
+        let ors = vec![ResultEvent::new(hashmap! {
+            sym!("resource") => term!(op!(And,
+                term!(op!(Isa, var!("_this"), term!(pattern!(instance!("Resource"))))),
+                term!(op!(Unify, term!(1), term!(op!(Dot, term!(op!(Dot, term!(op!(Dot, term!(op!(Dot, var!("_this"), str!("foo"))), str!("boo"))), str!("goo"))), str!("id")))))
+            ))
+        })];
+
+        let Filter { relations, .. } = Filter::build(types, ors, "resource", "Resource")?;
+        assert_eq!(
+            relations,
+            vec![
+                Relation(s("Resource"), s("foo"), s("Foo")),
+                Relation(s("Foo"), s("boo"), s("Boo")),
+                Relation(s("Boo"), s("goo"), s("Goo"))
+            ]
+        );
+
+        Ok(())
+    }
 }
