@@ -8,7 +8,6 @@ use crate::{
     error::RuntimeError,
     filter::singleton,
     folder::{fold_term, Folder},
-    formatting::ToPolarString,
     terms::*,
 };
 
@@ -74,12 +73,7 @@ impl Folder for VariableSubber {
 
 /// Substitute `sym!("_this")` for a variable in a partial.
 pub fn sub_this(this: Symbol, term: Term) -> Term {
-    if term
-        .value()
-        .as_symbol()
-        .map(|s| s == &this)
-        .unwrap_or(false)
-    {
+    if term.as_symbol().map(|s| s == &this).unwrap_or(false) {
         return term;
     }
     fold_term(term, &mut VariableSubber::new(this))
@@ -124,7 +118,7 @@ pub fn simplify_partial(
     simplify_debug!("*** simplify partial {:?}", var);
     simplifier.simplify_partial(&mut term);
     term = simplify_trivial_constraint(var.clone(), term);
-    simplify_debug!("simplify partial done {:?}, {:?}", var, term.to_polar());
+    simplify_debug!("simplify partial done {:?}, {}", var, term);
     if matches!(term.value(), Value::Expression(e) if e.operator != Operator::And) {
         (op!(And, term).into(), simplifier.perf_counters())
     } else {
@@ -148,7 +142,7 @@ pub fn simplify_bindings_opt(bindings: Bindings, all: bool) -> Result<Option<Bin
     if_debug! {
         eprintln!("before simplified");
         for (k, v) in bindings.iter() {
-            eprintln!("{:?} {:?}", k, v.to_polar());
+            eprintln!("{:?} {}", k, v);
         }
     }
 
@@ -171,7 +165,7 @@ pub fn simplify_bindings_opt(bindings: Bindings, all: bool) -> Result<Option<Bin
                 perf.merge(p);
             }
 
-            match simplified.value().as_expression() {
+            match simplified.as_expression() {
                 Ok(o) if o == &FALSE => unsatisfiable = true,
                 _ => (),
             }
@@ -213,7 +207,7 @@ pub fn simplify_bindings_opt(bindings: Bindings, all: bool) -> Result<Option<Bin
         if_debug! {
             eprintln!("after simplified");
             for (k, v) in simplified_bindings.iter() {
-                eprintln!("{:?} {:?}", k, v.to_polar());
+                eprintln!("{:?} {}", k, v);
             }
         }
 
@@ -238,13 +232,13 @@ impl fmt::Display for PerfCounters {
         writeln!(f, "perf {{")?;
         writeln!(f, "simplify term")?;
         for (term, ncalls) in self.simplify_term.iter() {
-            writeln!(f, "\t{}: {}", term.to_polar(), ncalls)?;
+            writeln!(f, "\t{}: {}", term, ncalls)?;
         }
 
         writeln!(f, "preprocess and")?;
 
         for (term, ncalls) in self.preprocess_and.iter() {
-            writeln!(f, "\t{}: {}", term.to_polar(), ncalls)?;
+            writeln!(f, "\t{}: {}", term, ncalls)?;
         }
 
         writeln!(f, "}}")
@@ -435,7 +429,7 @@ impl Simplifier {
     ) {
         fn toss_trivial_unifies(args: &mut TermList) {
             args.retain(|c| {
-                let o = c.value().as_expression().unwrap();
+                let o = c.as_expression().unwrap();
                 match o.operator {
                     Operator::Unify | Operator::Eq => {
                         assert_eq!(o.args.len(), 2);
@@ -472,24 +466,19 @@ impl Simplifier {
                 let mut keep = o.args.iter().map(|_| true).collect::<Vec<bool>>();
                 let mut references = o.args.iter().map(|_| false).collect::<Vec<bool>>();
                 for (i, arg) in o.args.iter().enumerate() {
-                    match self.maybe_bind_constraint(arg.value().as_expression().unwrap()) {
+                    match self.maybe_bind_constraint(arg.as_expression().unwrap()) {
                         MaybeDrop::Keep => (),
                         MaybeDrop::Drop => keep[i] = false,
                         MaybeDrop::Bind(var, value) => {
                             keep[i] = false;
-                            simplify_debug!("bind {:?}, {:?}", var, value.to_polar());
+                            simplify_debug!("bind {:?}, {}", var, value);
                             self.bind(var, value);
                         }
                         MaybeDrop::Check(var, value) => {
-                            simplify_debug!("check {:?}, {:?}", var.to_polar(), value.to_polar());
+                            simplify_debug!("check {}, {}", var, value);
                             for (j, arg) in o.args.iter().enumerate() {
                                 if j != i && arg.contains_variable(&var) {
-                                    simplify_debug!(
-                                        "check bind {:?}, {:?} ref: {}",
-                                        var.to_polar(),
-                                        value.to_polar(),
-                                        j
-                                    );
+                                    simplify_debug!("check bind {}, {} ref: {}", var, value, j);
                                     self.bind(var, value);
                                     keep[i] = false;
 
@@ -524,7 +513,6 @@ impl Simplifier {
                 simplifier.simplify_partial(&mut simplified);
                 *o = invert_operation(
                     simplified
-                        .value()
                         .as_expression()
                         .expect("a simplified expression")
                         .clone(),
@@ -548,7 +536,7 @@ impl Simplifier {
             // to avoid cloning to insert terms.
             let mut seen: HashSet<u64> = HashSet::with_capacity(args.len());
             args.retain(|a| {
-                let o = a.value().as_expression().unwrap();
+                let o = a.as_expression().unwrap();
                 o != &TRUE // trivial
                     && !seen.contains(&Term::from(o.mirror()).hash_value()) // reflection
                     && seen.insert(a.hash_value()) // duplicate
@@ -630,7 +618,7 @@ impl Simplifier {
             _ => (),
         }
 
-        if let Ok(sym) = orig.value().as_symbol() {
+        if let Ok(sym) = orig.as_symbol() {
             if term.contains_variable(sym) {
                 *term = orig.clone()
             }
@@ -644,7 +632,7 @@ impl Simplifier {
         let mut last = term.hash_value();
         let mut nbindings = self.bindings.len();
         loop {
-            simplify_debug!("simplify loop {:?}", term.to_polar());
+            simplify_debug!("simplify loop {}", term);
             self.counters.simplify_term();
 
             self.simplify_term(term, Simplifier::simplify_operation_variables);
